from vivarium_inputs.__about__ import (
    __author__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __uri__,
)
from vivarium_inputs._version import __version__
from vivarium_inputs.interface import (
    get_age_bins,
    get_demographic_dimensions,
    get_measure,
    get_population_structure,
    get_raw_data,
    get_theoretical_minimum_risk_life_expectancy,
)
